import bpy
import numpy as np

from blender_plots import blender_utils as bu


def get_frustum(
    intrinsics,
    height,
    width,
    image_depth,
    name="",
    with_fill=True,
    thickness=0.03,
    color=None,
    color_fill=None,
):
    frustum_points = (
        np.array([[0, height, 1], [width, height, 1], [0, 0, 1], [width, 0, 1]])
        * image_depth
    )

    frustum_points = np.einsum(
        "ij,...j->...i",
        np.linalg.inv(intrinsics),
        frustum_points,
    )
    frustum_edges = np.array(
        [[0, 1], [1, 3], [3, 2], [2, 0], [0, 4], [1, 4], [2, 4], [3, 4]]
    )

    frustum_faces = [
        [0, 1, 4],
        [1, 3, 4],
        [3, 2, 4],
        [2, 0, 4],
    ]

    collection = bu.new_collection(name) if with_fill else None
    mesh = bpy.data.meshes.new("frustum")
    mesh.from_pydata(
        np.vstack([frustum_points, np.zeros(3)]), frustum_edges, frustum_faces
    )

    frustum = bu.new_empty(f"{name}_frustum", mesh, collection=collection)
    modifier = bu.add_modifier(
        frustum,
        "WIREFRAME",
        use_crease=True,
        crease_weight=0.6,
        thickness=thickness,
        use_boundary=True,
    )
    bpy.ops.object.modifier_apply(modifier=modifier.name)

    if color is not None:
        bu.add_mesh_color(frustum.data, color)

    if with_fill:
        mesh_fill = bpy.data.meshes.new("fill")
        mesh_fill.from_pydata(
            np.vstack([frustum_points, np.zeros(3)]),
            frustum_edges,
            frustum_faces + [[0, 1, 3, 2]],
        )
        bu.new_empty(f"{name}_fill", mesh_fill, collection=collection)
        if color_fill is not None:
            bu.add_mesh_color(mesh_fill, color_fill)
        elif color is not None:
            bu.add_mesh_color(mesh_fill, color)
        return collection
    else:
        return frustum


def get_rotaitons_facing_point(origin, points):
    n_points = len(points)
    d = (origin - points) / np.linalg.norm(origin - points, axis=-1)[:, None]
    R = np.zeros((n_points, 3, 3))
    R[..., -1] = d
    R[..., 0] = np.cross(d, np.random.randn(n_points, 3))
    R[..., 0] /= np.linalg.norm(R[..., 0], axis=-1)[..., None]
    R[..., 1] = np.cross(R[..., 2], R[..., 0])
    return R



from blender_plots import (
    arrows,
    blender_utils,
    marker_utils,
    plots_base,
    scatter,
	scene_utils,
    surface,
)
from blender_plots.arrows import Arrows
from blender_plots.scatter import Scatter
from blender_plots.surface import Surface



import math

import bpy
import numpy as np

import blender_plots as bplt
from blender_plots import blender_utils as bu


def create_camera(location, rotation):
    if "Camera" in bpy.data.objects:
        bpy.data.objects.remove(bpy.data.objects["Camera"])

    bpy.ops.object.camera_add(
        enter_editmode=False, align="VIEW", location=location, rotation=rotation
    )
    bpy.context.scene.camera = bpy.data.objects["Camera"]


def render_image(output_path, resolution=None, samples=100):
    if resolution is not None:
        bpy.context.scene.render.resolution_x = resolution[0]
        bpy.context.scene.render.resolution_y = resolution[1]
    bpy.context.scene.cycles.samples = samples
    bpy.context.scene.render.filepath = output_path
    bpy.ops.render.render(write_still=True)


def setup_scene(
    clear=False,
    camera_location=None,
    camera_rotation=None,
    resolution=None,
    sun_energy=10,
):
    """Example function for setting up a scene for rendering."""
    if camera_location is None:
        camera_location = np.array([0, -5.321560, 2.042498]) * 0.6
    if camera_rotation is None:
        camera_rotation = [math.radians(68.4), 0.0, 0.0]

    if clear:
        bpy.ops.wm.read_homefile()
    bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[0].default_value = (
        1,
        1,
        1,
        1,
    )
    bpy.context.scene.render.engine = "CYCLES"
    bpy.data.scenes["Scene"].cycles.samples = 256

    if "Light" in bpy.data.objects:
        # remove default light
        bpy.data.objects.remove(bpy.data.objects["Light"])
    if "Sun" in bpy.data.objects:
        bpy.data.objects.remove(bpy.data.objects["Sun"])
    bpy.ops.object.light_add(
        type="SUN", radius=1, align="WORLD", location=(0, 0, 0), scale=(1, 1, 1)
    )
    bpy.data.objects["Sun"].data.energy = sun_energy
    bpy.data.objects["Sun"].data.angle = np.pi / 2
    bpy.data.worlds["World"].node_tree.nodes["Background"].inputs[
        "Strength"
    ].default_value = 0.5

    bpy.context.scene.render.film_transparent = True
    create_camera(camera_location, camera_rotation)
    if resolution is not None:
        bpy.context.scene.render.resolution_x = resolution[0]
        bpy.context.scene.render.resolution_y = resolution[1]



import numpy as np

from blender_plots import blender_utils as bu
from blender_plots import plots_base, scatter


class Arrows(plots_base.Plot):
    """Create an arrow plot.

    Args:
        start: xyz coordinate(s) of arrow starts (3, Nx3 or TxNx3 array - see scatter.py)
        vector: xyz components of arrow vectors, same size as `start`
        end: xyz coordinate(s) of arrow ends (can be provided instead of `vector`)
        color: (Tx)Nx3 or (Tx)Nx4 array or with RGB or RGBA values for each point and (optionally) time,
            or a single RGB/RGBA-value (e.g. (1, 0, 0) for red) to apply to every point.
        name: name to use for blender object. Will delete any previous plot with the same name.
        head_length: length of the arrow head.
        radius: radius of the arrow stem.
        radius_ratio: ratio of the arrow head radius to the stem radius.
        end_trim_length: length by which to reduce each arrow. If set to 0 the arrow tip will be exactly at `start`+`vector` (or `end` if provided)
    """

    def __init__(
        self,
        start,
        vector=None,
        end=None,
        color=None,
        name="arrows",
        head_length=1,
        radius=1,
        radius_ratio=1.5,
        end_trim_length=0,
    ):
        super(Arrows, self).__init__(
            start, None, None, color=color, name=name, n_dims=1
        )

        self.node_linker = add_arrows(
            self.modifier,
            self.n_frames,
            head_length,
            radius,
            radius_ratio,
            end_trim_length,
        )

        self.modifier[self.node_linker.input_sockets["Point Color"]] = (
            self.color_material
        )

        if (end is not None) and (vector is not None):
            raise ValueError(
                "Cannot provide both 'end' and 'vector' as input arguments."
            )
        elif end is not None:
            self.arrows = end - start
        else:
            self.arrows = vector
        self.base_object.data.update()

    def get_geometry(self):
        return self._points.reshape(-1, 3), [], []

    @property
    def arrows(self):
        return self._arrows

    @arrows.setter
    def arrows(self, arrows):
        self._arrows = np.array(arrows) if arrows is not None else arrows
        self.update_arrows()

    def update_arrows(self):
        if self._arrows is not None:
            arrows, marker_dims = self.tile_data(
                self._arrows, [[3], []], "marker scale"
            )
            if marker_dims == []:
                arrows = np.array([arrows] * 3).T
            bu.set_vertex_attribute(
                self.mesh, bu.Constants.ARROWS, arrows, "FLOAT_VECTOR"
            )


def add_arrows(
    base_modifier, n_frames, head_length, radius, radius_ratio, end_trim_length
):
    """ """
    node_linker = bu.get_node_linker(base_modifier)
    node_linker.new_input_socket("Point Color", "NodeSocketMaterial")
    if n_frames is None:
        frame_selection = None
    else:
        frame_selection = bu.get_frame_selection_node(base_modifier, n_frames).outputs[
            "Value"
        ]

    points_socket = node_linker.new_node(
        "GeometryNodeMeshToPoints", mesh=node_linker.group_input.outputs["Geometry"]
    ).outputs["Points"]

    arrows = node_linker.new_node(
        "GeometryNodeInputNamedAttribute",
        data_type="FLOAT_VECTOR",
        name=bu.Constants.ARROWS,
    ).outputs["Attribute"]
    lengths = node_linker.new_node(
        "ShaderNodeVectorMath", operation="LENGTH", vector=arrows
    ).outputs["Value"]
    lengths = node_linker.new_node(
        "ShaderNodeMath",
        operation="SUBTRACT",
        input_0=lengths,
        input_1=head_length + end_trim_length,
    ).outputs["Value"]
    rotations = node_linker.new_node(
        "FunctionNodeAlignEulerToVector", axis="Z", vector=arrows
    ).outputs["Rotation"]

    cylinder_mesh = node_linker.new_node(
        "GeometryNodeMeshCylinder", depth=1, radius=radius
    ).outputs["Mesh"]
    cylinder_mesh = node_linker.new_node(
        "GeometryNodeTransform", geometry=cylinder_mesh, translation=[0, 0, 0.5]
    ).outputs["Geometry"]
    colored_cylinder = node_linker.new_node(
        "GeometryNodeSetMaterial",
        geometry=cylinder_mesh,
        material=node_linker.group_input.outputs["Point Color"],
    ).outputs["Geometry"]
    cylinder_instances = node_linker.new_node(
        "GeometryNodeInstanceOnPoints",
        points=points_socket,
        selection=frame_selection,
        instance=colored_cylinder,
        rotation=rotations,
        scale=node_linker.new_node("ShaderNodeCombineXYZ", x=1, y=1, z=lengths).outputs[
            "Vector"
        ],
    ).outputs["Instances"]
    cylinder_geometry = node_linker.new_node(
        "GeometryNodeRealizeInstances", geometry=cylinder_instances
    ).outputs["Geometry"]

    cone_mesh = node_linker.new_node(
        "GeometryNodeMeshCone", depth=head_length, radius_bottom=radius * radius_ratio
    ).outputs["Mesh"]
    colored_cone = node_linker.new_node(
        "GeometryNodeSetMaterial",
        geometry=cone_mesh,
        material=node_linker.group_input.outputs["Point Color"],
    ).outputs["Geometry"]
    cone_instances = node_linker.new_node(
        "GeometryNodeInstanceOnPoints",
        points=points_socket,
        selection=frame_selection,
        instance=colored_cone,
        rotation=rotations,
    ).outputs["Instances"]
    cone_instances = node_linker.new_node(
        "GeometryNodeTranslateInstances",
        instances=cone_instances,
        translation=node_linker.new_node(
            "ShaderNodeCombineXYZ", x=0, y=0, z=lengths
        ).outputs["Vector"],
    ).outputs["Instances"]
    cone_geometry = node_linker.new_node(
        "GeometryNodeRealizeInstances", geometry=cone_instances
    ).outputs["Geometry"]
    arrow_geometry_node = node_linker.new_node(
        "GeometryNodeJoinGeometry", geometry=cylinder_geometry
    )
    node_linker.node_group.links.new(cone_geometry, arrow_geometry_node.inputs[0])

    node_linker.new_node(
        "NodeGroupOutput", geometry=arrow_geometry_node.outputs["Geometry"]
    )
    return node_linker



from setuptools import setup

setup(
    name='blender-plots',
    version='1.0.12',
    author='Linus Härenstam-Nielsen',
    description='Plotting API for Blender',
    url='https://github.com/Linusnie/blender-plots',
    python_requires='>=3.10',
    packages=['blender_plots'],
)